"""Script to run parse conformance on all package versions in Package Hub.

Run as a CLI: `uv run package-hub-compat`

Reads JSON from a local clone of `hub.getdbt.com` and extracts all package versions,
then runs parse conformance using `check_parse_conformance`.
Writes output to `output/conformance_output.json`, which is used in two other scripts
* `update_package_hub_json`: adds compatibility info back to the original Package Hub files
* `conformance_output`: produces 2 CSV files that summarize compatibility info for further analysis
"""

import json
import os
import re
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
import typer
from rich import print
from rich.console import Console
from typing_extensions import Annotated

from dbt_fusion_package_tools.check_parse_conformance import (
    download_tarball_and_run_conformance,
)
from dbt_fusion_package_tools.compatibility import FusionConformanceResult

console = Console()
error_console = Console(stderr=True)

app = typer.Typer()

current_dir = Path.cwd()
DEFAULT_OUTPUT_PATH = current_dir / "src" / "dbt_fusion_package_tools" / "scripts" / "output"
DEFAULT_HUB_PATH = Path.home() / "workplace" / "hub.getdbt.com"
DEFAULT_FUSION_BINARY_PATH = Path.home() / ".local" / "bin" / "dbt"


# Example package index path:
# data/packages/Aaron-Zhou/synapse_statistic/index.json
def is_package_index_file(file_path: str) -> bool:
    file_path_split = file_path.split("/")
    if len(file_path_split) != 5:
        return False
    return file_path_split[-1] == "index.json"


# Example package version path:
# data/packages/Aaron-Zhou/synapse_statistic/versions/v0.1.0.json
def is_package_version_file(file_path: str) -> bool:
    file_path_split = file_path.split("/")
    if len(file_path_split) != 6:
        return False
    return file_path_split[-2] == "versions"


# Example paths that resolve to Aaron-Zhou/synapse_statistic
# data/packages/Aaron-Zhou/synapse_statistic/index.json
# data/packages/Aaron-Zhou/synapse_statistic/versions/v0.1.0.json
def extract_package_id_from_path(file_path: str) -> str:
    file_path_split = file_path.split("/")
    if file_path_split[0] != "data" or file_path_split[1] != "packages" or len(file_path_split) < 4:
        return ""
    return f"{file_path_split[2]}/{file_path_split[3]}"


def clean_version(version_str: Optional[str]) -> str:
    """Remove leading 'v' or 'V' from version strings, if present."""
    if version_str is None:
        return ""
    elif version_str.startswith(("v", "V")):
        return version_str[1:]
    return version_str


# Notes:
# The package_id from path (which is the key in results)
# is the organization + package name
# This is not related to the Github URL
# Example:
# package_id_from_path = AxelThevenot/dbt_star
# package_version_download_url = https://codeload.github.com/AxelThevenot/dbt-star/tar.gz/0.1.0
# Package Hub page: https://hub.getdbt.com/AxelThevenot/dbt_star/latest/
# package name in packages.yml: AxelThevenot/dbt_star
def process_json(file_path: str, parsed_json: Any) -> dict[str, Any]:
    package_id = extract_package_id_from_path(file_path)
    if package_id == "":
        return {}
    if is_package_index_file(file_path):
        return {
            "package_id_from_path": package_id,
            "package_latest_version_index_json": clean_version(parsed_json.get("latest")),
            "package_name_index_json": parsed_json.get("name"),
            "package_namespace_index_json": parsed_json.get("namespace"),
            "package_redirect_name": parsed_json.get("redirectname"),
            "package_redirect_namespace": parsed_json.get("redirectnamespace"),
        }
    elif is_package_version_file(file_path):
        if "_source" in parsed_json:
            github_url = parsed_json["_source"].get("url", "")
        else:
            github_url = None
        if "downloads" in parsed_json:
            tarball_url = parsed_json["downloads"].get("tarball")
        else:
            tarball_url = None
        return {
            "package_id_from_path": package_id,
            "package_id_with_version": parsed_json.get("id"),
            "package_name_version_json": parsed_json.get("name"),
            "package_version_string": clean_version(parsed_json.get("version")),
            "package_version_require_dbt_version": parsed_json.get("require_dbt_version"),
            "package_version_github_url": github_url,
            "package_version_download_url": tarball_url,
        }
    else:
        return {}


def write_dict_to_json(data: Dict[str, Any], dest_dir: Path, *, indent: int = 2, sort_keys: bool = True) -> None:
    out_file = dest_dir / "package_output.json"
    with out_file.open("w", encoding="utf-8") as fh:
        json.dump(data, fh, indent=indent, sort_keys=sort_keys, ensure_ascii=False)


def read_json_from_local_hub_repo(path: str, file_count_limit: int = 0):
    """Read JSON files from a local copy of the hub repo and return a
    defaultdict mapping package_id -> list[parsed outputs].

    The `path` argument may be either:
      - the repository root (so files are found under `data/packages/...`),
      - or the `data/packages` directory itself, or
      - a single JSON file path.

    Behavior mirrors `download_package_jsons_from_hub_repo` where possible:
      - JSON files are found recursively
      - each file is parsed and passed to `process_json(file_path, parsed_json)`
      - parsing/IO errors are warned and skipped
    """
    base = Path(path)
    packages: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)

    if not base.exists():
        warnings.warn(f"Path does not exist: {path}")
        return packages

    # Collect JSON files
    json_files: List[Path]
    if base.is_file():
        if base.suffix.lower() == ".json":
            json_files = [base]
        else:
            return packages
    else:
        json_files = sorted(base.rglob("*.json"), key=lambda p: str(p))

    if file_count_limit > 0:
        json_files = json_files[:file_count_limit]

    if not json_files:
        return packages

    for file in json_files:
        try:
            with file.open("r", encoding="utf-8") as fh:
                parsed = json.load(fh)

            # Try to produce a repo-style path like 'data/packages/...'
            file_path: str
            parts = list(file.parts)
            if "data" in parts:
                idx = parts.index("data")
                file_path = Path(*parts[idx:]).as_posix()
            else:
                try:
                    # prefer path relative to provided base
                    rel = file.relative_to(base)
                    file_path = rel.as_posix()
                except Exception:
                    file_path = file.as_posix()

            # If the user passed the `data/packages` directory itself,
            # ensure returned path still starts with 'data/packages'
            if not file_path.startswith("data/packages") and base.name == "packages" and base.parent.name == "data":
                rel = file.relative_to(base)
                file_path = Path("data") / "packages" / rel
                file_path = file_path.as_posix()

            output = process_json(file_path, parsed)
            if output != {}:
                packages[output["package_id_from_path"]].append(output)
        except Exception as exc:
            warnings.warn(f"Failed to read/parse {file}: {exc}")

    return packages


def reload_packages_from_file(
    file_path: Path,
) -> defaultdict[str, list[dict[str, Any]]]:
    with file_path.open("r", encoding="utf-8") as fh:
        return json.load(fh)


def get_github_repos_from_file(file_path: Path) -> defaultdict[str, set[str]]:
    """Extracts the Github repo paths from package hub output.

    Note that the repos are in a set because in rare cases, a package
    may have more than 1 reported repo.

    Args:
        file_path (Path): path to output from read_json_from_local_hub_repo

    Returns:
        defaultdict[str, set[str]]: packages with repos

    Example:
        {
            "package_1":
                {"https://github.com/example/package-1"}
        }
    """
    with file_path.open("r", encoding="utf-8") as fh:
        output = json.load(fh)
    package_repos: defaultdict[str, set[str]] = defaultdict(set)
    for package in output:
        for version in output[package]:
            repo = version.get("package_version_github_url", "")
            m = re.match(r"^(https?://github\.com/[^/]+/[^/]+)", repo)
            if m:
                package_repos[package].add(m.group(1))
    for repos in package_repos:
        repo_count = len(package_repos[repos])
        if repo_count > 1:
            print(repos, package_repos[repos])
    return package_repos


def check_github_url(
    url: str, timeout: int = 10, github_token: Optional[str] = os.getenv("GITHUB_TOKEN")
) -> Dict[str, Any]:
    """Check a GitHub URL and return status info.
    Returns a dict with keys: status (int|None), is_404 (bool), is_301 (bool),
    location (redirect target or None), error (str|None).
    """
    headers = {"User-Agent": "dbt-autofix-agent"}
    if github_token:
        headers["Authorization"] = f"token {github_token}"

    try:
        resp = requests.head(url, headers=headers, timeout=timeout, allow_redirects=False)
        if resp.status_code in (405, 501):  # HEAD not allowed => try GET
            resp = requests.get(
                url,
                headers=headers,
                timeout=timeout,
                allow_redirects=False,
                stream=True,
            )
        status = resp.status_code
        location = resp.headers.get("Location")
        return {
            "status": status,
            "is_404": status == 404,
            "is_301": status == 301,
            "location": location,
            "error": None,
        }
    except requests.RequestException as exc:
        return {
            "status": None,
            "is_404": False,
            "is_301": False,
            "location": None,
            "error": str(exc),
        }


def validate_github_urls(packages: defaultdict[str, set[str]], package_limit: int = 0) -> dict[str, str]:
    # Returning a single string here because literally only 1 package has more than 1 valid repo
    # and it looks like a mistake (Saras-Daton/Walmart)
    valid_repos: dict[str, str] = {}
    for i, package in enumerate(packages):
        if package_limit > 0 and i > package_limit:
            break
        for github_url in packages[package]:
            response = check_github_url(github_url)
            if not response:
                error_console.log(f"No response for {package} {github_url}")
            if (response["status"] != 200 and not response["is_301"]) or response.get("error"):
                error_console.log(response)
            if response["is_404"]:
                continue
            elif response["is_301"]:
                valid_repos[package] = response["location"]
            else:
                valid_repos[package] = github_url
    return valid_repos


def follow_redirects(package_name: str, packages: dict[str, dict[str, Optional[str]]]) -> str:
    package_redirect_name: Optional[str] = packages[package_name].get("package_redirect_name")
    package_redirect_namespace: Optional[str] = packages[package_name].get("package_redirect_namespace")
    # base case: package does not have any redirects
    if not package_redirect_name and not package_redirect_namespace:
        return package_name
    # recursive case: follow redirect
    original_namespace = package_name.split("/")[0]
    original_name = package_name.split("/")[1]
    if package_redirect_name and package_redirect_namespace:
        package_after_redirect: str = f"{package_redirect_namespace}/{package_redirect_name}"
    elif package_redirect_name:
        package_after_redirect: str = f"{original_namespace}/{package_redirect_name}"
    else:
        package_after_redirect: str = f"{package_redirect_namespace}/{original_name}"
    next_redirect_name = follow_redirects(package_after_redirect, packages)
    return next_redirect_name


def get_latest_github_tarball_urls(hub_data: defaultdict[str, list[dict[str, Any]]]) -> dict[str, str]:
    # first load in all packages and get latest version + redirects
    # but don't actually follow the redirects yet
    packages_no_redirects: dict[str, dict[str, Optional[str]]] = {}
    for package in hub_data:
        # index file should always be first
        package_latest_version: str = hub_data[package][0]["package_latest_version_index_json"]
        package_redirect_name: Optional[str] = hub_data[package][0].get("package_redirect_name")
        package_redirect_namespace: Optional[str] = hub_data[package][0].get("package_redirect_namespace")
        package_latest_version_download_url: Optional[str] = None

        for version in hub_data[package][1:]:
            package_version_string = version.get("package_version_string")

            # get the tarball url for the latest version only
            if package_version_string == package_latest_version:
                package_latest_version_download_url = version.get("package_version_download_url")
                break

        if not package_latest_version_download_url:
            console.log(f"No download available for {package}")
            continue
        else:
            packages_no_redirects[package] = {
                "package_latest_version": package_latest_version,
                "package_redirect_name": package_redirect_name,
                "package_redirect_namespace": package_redirect_namespace,
                "package_latest_version_download_url": package_latest_version_download_url,
            }

    # get final latest version url after following redirect
    package_latest_version_urls: dict[str, str] = {}
    for package in packages_no_redirects:
        package_latest_name: str = follow_redirects(package, packages_no_redirects)
        package_latest_url: Optional[str] = packages_no_redirects[package_latest_name].get(
            "package_latest_version_download_url"
        )
        if package_latest_url is not None:
            package_latest_version_urls[package] = package_latest_url
    return package_latest_version_urls


def run_conformance_from_tarballs(
    output: defaultdict[str, list[dict[str, Any]]],
    package_latest_version_urls: dict[str, str],
    package_limit: int = 0,
    fusion_binary=None,
) -> dict[str, dict[str, FusionConformanceResult]]:
    results: dict[str, dict[str, FusionConformanceResult]] = {}

    for i, package in enumerate(output):
        if package_limit > 0 and i > package_limit:
            break
        results[package] = {}
        for version in output[package]:
            package_version_download_url = version.get("package_version_download_url")
            package_version_string = version.get("package_version_string")
            if package_version_string is None:
                continue
            if package_version_download_url is None:
                console.log(f"No download URL found for {package} version {package_version_string}")
                continue
            conformance_output = download_tarball_and_run_conformance(
                package_name=package,
                package_id=version["package_id_from_path"],
                package_version_str=str(package_version_string),
                package_version_download_url=package_version_download_url,
                latest_package_version_download_url=package_latest_version_urls.get(package),
                fusion_binary=fusion_binary,
            )
            if not conformance_output:
                console.log(f"Could not run conformance for {package} version {package_version_string}\n")
                continue
            else:
                results[package][package_version_string] = conformance_output
                console.log()

    return results


def write_conformance_output_to_json(
    data: dict[str, dict[str, FusionConformanceResult]],
    dest_dir: Path,
    *,
    indent: int = 2,
    sort_keys: bool = True,
):
    data_output = {}
    for k, v in data.items():
        data_output[k] = {version: result.to_dict() for version, result in v.items()}
    out_file = Path(dest_dir) / "conformance_output.json"
    with out_file.open("w", encoding="utf-8") as fh:
        json.dump(data_output, fh, indent=indent, sort_keys=sort_keys, ensure_ascii=False)


@app.command()
def main(
    local_hub_path: Annotated[
        str, typer.Option("--local-hub", help="Fully qualified path to local Package Hub clone")
    ] = str(DEFAULT_HUB_PATH),
    fusion_binary: Annotated[str, typer.Option("--fusion-binary", help="Name of fusion binary")] = str(
        DEFAULT_FUSION_BINARY_PATH
    ),
    output_path: Annotated[
        str, typer.Option("--output-path", help="Fully qualified path to directory for output")
    ] = str(DEFAULT_OUTPUT_PATH),
    package_limit: Annotated[
        int, typer.Option("--limit", help="Only run on first n packages (default = 0 to run all packages)")
    ] = 0,
):
    if output_path:
        output_dir = Path(output_path)
    else:
        output_dir = DEFAULT_OUTPUT_PATH
    console.log(f"Reading from local Hub repo: {local_hub_path}")
    console.log(f"Writing to output path: {output_dir}/conformance_output.json")
    console.log(f"Package limit: {package_limit}")
    console.log(f"Fusion binary: {fusion_binary}")

    output: defaultdict[str, list[dict[str, Any]]] = read_json_from_local_hub_repo(path=local_hub_path)
    package_latest_version_urls: dict[str, str] = get_latest_github_tarball_urls(output)
    parse_conformance_results = run_conformance_from_tarballs(
        output, package_latest_version_urls, package_limit, fusion_binary
    )
    write_conformance_output_to_json(parse_conformance_results, output_path)
    console.log(f"Successfully wrote output to {output_path}/conformance_output.json")


if __name__ == "__main__":
    app()



from dataclasses import dataclass, field
from typing import Any, Optional

from rich.console import Console

from dbt_fusion_package_tools.compatibility import FusionConformanceResult
from dbt_fusion_package_tools.manual_overrides import (
    EXPLICIT_ALLOW_ALL_VERSIONS,
    EXPLICIT_DISALLOW_ALL_VERSIONS,
    EXPLICIT_DISALLOW_VERSIONS,
)
from dbt_fusion_package_tools.upgrade_status import PackageVersionFusionCompatibilityState
from dbt_fusion_package_tools.version_utils import (
    FUSION_COMPATIBLE_VERSION,
    VersionRange,
    VersionSpecifier,
    construct_version_list_from_raw,
    convert_version_specifiers_to_range,
    get_version_specifiers,
    versions_compatible,
)

console = Console()


@dataclass
class DbtPackageVersion:
    package_name: str
    package_version_str: str
    require_dbt_version_range: list[str] = field(default_factory=list)
    # version: VersionSpecifier = field(init=False)
    require_dbt_version: Optional[VersionRange] = field(init=False)
    package_id_with_version: Optional[str] = None
    package_id: Optional[str] = None
    raw_require_dbt_version_range: Any = None
    fusion_conformance_result: Optional[FusionConformanceResult] = None

    @property
    def version(self) -> VersionSpecifier:
        return VersionSpecifier.from_version_string(self.package_version_str)

    @version.setter
    def version(self, new_version: VersionSpecifier) -> None:
        self.package_version_str = new_version.to_version_string(skip_matcher=True)

    def __post_init__(self):
        try:
            if self.raw_require_dbt_version_range is not None:
                self.require_dbt_version_range = construct_version_list_from_raw(self.raw_require_dbt_version_range)
            if self.require_dbt_version_range and len(self.require_dbt_version_range) > 0:
                version_specs: list[VersionSpecifier] = get_version_specifiers(self.require_dbt_version_range)
                self.require_dbt_version = convert_version_specifiers_to_range(version_specs)
            else:
                self.require_dbt_version = None
        except:
            self.require_dbt_version = None

    def __lt__(self, other) -> bool:
        if self.package_name != other.package_name:
            return False
        return self.version < other.version

    def __eq__(self, other) -> bool:
        return self.package_name == other.package_name and self.version == other.version

    def is_prerelease_version(self) -> bool:
        return self.version.prerelease is not None

    def is_require_dbt_version_fusion_compatible(self) -> bool:
        if self.require_dbt_version:
            return versions_compatible(self.require_dbt_version, FUSION_COMPATIBLE_VERSION)
        else:
            return False

    def is_require_dbt_version_defined(self) -> bool:
        return self.require_dbt_version_range != None and len(self.require_dbt_version_range) > 0

    def is_version_explicitly_disallowed_on_fusion(self) -> bool:
        return (
            self.package_id is not None
            and self.package_id in EXPLICIT_DISALLOW_VERSIONS
            and self.package_version_str in EXPLICIT_DISALLOW_VERSIONS[self.package_id]
        )

    def is_explicitly_disallowed_on_fusion(self) -> bool:
        if self.package_id is not None:
            if self.package_id in EXPLICIT_DISALLOW_ALL_VERSIONS:
                return True
            elif self.is_version_explicitly_disallowed_on_fusion():
                return True
        return False

    def is_explicitly_allowed_on_fusion(self) -> bool:
        if self.package_id is not None and self.package_id in EXPLICIT_ALLOW_ALL_VERSIONS:
            return True
        return False

    def get_fusion_compatibility_state(self) -> PackageVersionFusionCompatibilityState:
        if self.is_explicitly_allowed_on_fusion():
            return PackageVersionFusionCompatibilityState.EXPLICIT_ALLOW
        elif self.is_explicitly_disallowed_on_fusion():
            return PackageVersionFusionCompatibilityState.EXPLICIT_DISALLOW
        elif not self.is_require_dbt_version_defined():
            return PackageVersionFusionCompatibilityState.NO_DBT_VERSION_RANGE
        elif self.is_require_dbt_version_fusion_compatible():
            return PackageVersionFusionCompatibilityState.DBT_VERSION_RANGE_INCLUDES_2_0
        elif not self.is_require_dbt_version_fusion_compatible():
            return PackageVersionFusionCompatibilityState.DBT_VERSION_RANGE_EXCLUDES_2_0
        else:
            return PackageVersionFusionCompatibilityState.UNKNOWN



"""Script to add parse conformance data to Package Hub.

Run as a CLI: `uv run update-package-hub`

Reads the `conformance_output.json` output from `package_hub_fusion_compatibility`.
Adds or updates the Fusion compatibility info for package versions in a local Hub repo.
"""

import json
from collections import defaultdict
from pathlib import Path
from typing import Any

import typer
from rich.console import Console
from typing_extensions import Annotated

from dbt_fusion_package_tools.compatibility import FusionConformanceResult
from dbt_fusion_package_tools.exceptions import SemverError
from dbt_fusion_package_tools.version_utils import VersionSpecifier

console = Console()
error_console = Console(stderr=True)

app = typer.Typer()

current_dir = Path.cwd()
DEFAULT_OUTPUT_PATH = current_dir / "src" / "dbt_fusion_package_tools" / "scripts" / "output"
DEFAULT_HUB_PATH = Path.home() / "workplace" / "hub.getdbt.com"
DEFAULT_FUSION_BINARY_PATH = Path.home() / ".local" / "bin" / "dbt"
DEFAULT_CONFORMANCE_OUTPUT_FILE = (
    current_dir / "src" / "dbt_fusion_package_tools" / "scripts" / "output" / "conformance_output.json"
)


def reload_output_from_file(
    file_path: Path,
) -> defaultdict[str, dict[str, Any]]:
    with file_path.open("r", encoding="utf-8") as fh:
        return json.load(fh)


def extract_output_from_json(
    data: defaultdict[str, dict[str, Any]],
) -> dict[str, dict[str, FusionConformanceResult]]:
    output: dict[str, dict[str, FusionConformanceResult]] = {}
    for package in data:
        output[package] = {}
        for version_id in data[package]:
            version_data: dict[str, Any] = data[package][version_id]
            version_data_conformance_result = FusionConformanceResult.from_dict(version_data)
            output[package][version_id] = version_data_conformance_result
    return output


def check_for_rename(hub_path: str, package_name: str) -> VersionSpecifier:
    dir_path = Path(hub_path) / "data" / "packages" / package_name
    file_path = dir_path / "index.json"
    with file_path.open("r", encoding="utf-8") as fh:
        index_json = json.load(fh)
    # if "redirectname" in index_json or "redirectnamespace" in index_json:
    #     print(f"package {package_name} renamed after version {index_json['latest']}")
    return VersionSpecifier.from_version_string(index_json["latest"])


def find_package_hub_file(hub_path: str, package_name: str, version: str) -> Path:
    dir_path = Path(hub_path) / "data" / "packages" / package_name / "versions"
    if (dir_path / f"{version}.json").is_file():
        file_path = dir_path / f"{version}.json"
    else:
        file_path = dir_path / f"v{version}.json"
    return file_path


def get_json_from_package_hub_file(file_path: Path, package_name: str, version: str) -> dict[str, Any]:
    try:
        with file_path.open("r", encoding="utf-8") as fh:
            return json.load(fh)
    except FileNotFoundError:
        error_console.log(f"No package hub output found for {package_name} version {version}")
        return {}


def update_hub_json(
    original_json: dict[str, Any],
    conformance_output: FusionConformanceResult,
    fusion_version: str,
) -> dict[str, Any]:
    # this is done to ensure that the ordering matches the original so the git diff is minimized
    updated_json = {
        "id": original_json["id"],
        "name": original_json["name"],
        "version": original_json["version"],
        "published_at": original_json["published_at"],
        "packages": original_json["packages"],
    }
    if "require_dbt_version" in original_json:
        updated_json["require_dbt_version"] = original_json["require_dbt_version"]
    updated_json["works_with"] = original_json["works_with"]
    updated_json["_source"] = original_json["_source"]
    updated_json["downloads"] = original_json["downloads"]

    new_conformance = conformance_output.to_dict()
    if "fusion_compatibility" in original_json:
        manually_verified_compatible = original_json["fusion_compatibility"].get("manually_verified_compatible")
        manually_verified_incompatible = original_json["fusion_compatibility"].get("manually_verified_incompatible")
        if manually_verified_compatible:
            new_conformance["manually_verified_compatible"] = manually_verified_compatible
        if manually_verified_incompatible:
            new_conformance["manually_verified_incompatible"] = manually_verified_incompatible
    updated_json["fusion_compatibility"] = new_conformance
    return updated_json


def write_dict_to_json(data: dict[str, Any], dest_path: Path, *, indent: int = 2, sort_keys: bool = True) -> None:
    out_file = dest_path
    with out_file.open("w", encoding="utf-8") as fh:
        json.dump(data, fh, indent=indent, sort_keys=False, ensure_ascii=False)


@app.command()
def main(
    local_hub_path: Annotated[
        str, typer.Option("--local-hub", help="Fully qualified path to local Package Hub clone")
    ] = str(DEFAULT_HUB_PATH),
    fusion_binary: Annotated[str, typer.Option("--fusion-binary", help="Name of fusion binary")] = str(
        DEFAULT_FUSION_BINARY_PATH
    ),
    output_path: Annotated[
        str, typer.Option("--output-path", help="Fully qualified path to directory for output")
    ] = str(DEFAULT_OUTPUT_PATH),
    package_limit: Annotated[
        int, typer.Option("--limit", help="Only run on first n packages (default = 0 to run all packages)")
    ] = 0,
    fusion_version: Annotated[str, typer.Option("--fusion-version", help="Version of Fusion used for testing")] = "89",
    conformance_output_path: Annotated[
        str, typer.Option("--conformance-output", help="Path to conformance output")
    ] = str(DEFAULT_CONFORMANCE_OUTPUT_FILE),
):
    file_path: Path = Path(conformance_output_path)
    console.log(f"Writing to local Hub repo: {local_hub_path}")
    console.log(f"Reading from output path: {conformance_output_path}")
    console.log(f"Package limit: {str(package_limit) if package_limit > 0 else 'none'}")
    console.log(f"Fusion version: 2.0.0-preview.{fusion_version}")

    conformance_output = reload_output_from_file(file_path)
    conformance_data = extract_output_from_json(conformance_output)
    package_count = 0
    success_count = 0
    error_count = 0
    i = 0
    for package_name in conformance_data:
        if package_limit > 0 and i > package_limit:
            break
        console.log(f"Updating package {package_name} with {len(conformance_data[package_name])} versions")
        latest_version = check_for_rename(local_hub_path, package_name)
        package_count += 1
        for version in conformance_data[package_name]:
            try:
                version_spec = VersionSpecifier.from_version_string(version)
                if version_spec > latest_version:
                    continue
            except SemverError:
                error_console.log(f"Can't parse version spec {version} for package {package_name}")
                error_count += 1
                continue
            version_file_path = find_package_hub_file(local_hub_path, package_name, version)
            package_hub_json = get_json_from_package_hub_file(version_file_path, package_name, version)
            if package_hub_json == {}:
                error_count += 1
                continue
            updated_json = update_hub_json(
                package_hub_json,
                conformance_data[package_name][version],
                f"2.0.0-preview.{fusion_version}",
            )
            write_dict_to_json(updated_json, version_file_path, indent=4)
            success_count += 1
        i += 1
    console.log("Package Hub update complete", style="green")
    console.log(f"Packages processed: {package_count}")
    console.log(f"Success count: {success_count}")
    console.log(f"Error count: {error_count}")


if __name__ == "__main__":
    app()



import json
from pathlib import Path
from typing import Any, Optional

from dbt_fusion_package_tools.dbt_package_version import DbtPackageVersion
from dbt_fusion_package_tools.version_utils import VersionSpecifier


def read_package_output_json(file_path: Path) -> dict[str, Any]:
    """Read a JSON file containing package output data.

    Args:
        file_path: Path to the JSON file.

    Returns:
        A dictionary with the contents of the JSON file.
    """
    with file_path.open("r", encoding="utf-8") as f:
        data = json.load(f)

    return data


def convert_version_spec_to_string(
    version_spec: Optional[VersionSpecifier],
) -> Optional[str]:
    if version_spec is None:
        return None
    else:
        return version_spec.to_version_string()


def new_name_from_redirect(redirect_name, redirect_namespace, current_name, current_namespace) -> str:
    if redirect_name and redirect_namespace:
        return f"{redirect_namespace}/{redirect_name}"
    elif redirect_namespace == None:
        return f"{current_namespace}/{redirect_name}"
    else:
        return f"{redirect_namespace}/{current_name}"


def get_versions_for_package(package_versions) -> dict[str, Any]:
    versions: list[DbtPackageVersion] = []
    latest_fusion_version: Optional[VersionSpecifier] = None
    latest_version: Optional[VersionSpecifier] = None
    latest_version_incl_prerelease: Optional[VersionSpecifier] = None
    oldest_fusion_compatible_version: Optional[VersionSpecifier] = None
    fusion_compatible_versions: list[VersionSpecifier] = []
    fusion_incompatible_versions: list[VersionSpecifier] = []
    unknown_compatibility_versions: list[VersionSpecifier] = []
    package_latest_version_index_json: Optional[VersionSpecifier] = None
    package_redirect_name: Optional[str] = None
    package_redirect_namespace: Optional[str] = None
    for version in package_versions:
        # skip record for package index
        if "package_latest_version_index_json" in version:
            package_latest_version_index_json = VersionSpecifier.from_version_string(
                version["package_latest_version_index_json"]
            )
            if version["package_redirect_name"] != None or version["package_redirect_namespace"] != None:
                package_redirect_name = version["package_redirect_name"]
                package_redirect_namespace = version["package_redirect_namespace"]
            continue
        package_version = DbtPackageVersion(
            package_name=version["package_name_version_json"],
            package_version_str=version["package_version_string"],
            package_id_with_version=version["package_id_with_version"],
            raw_require_dbt_version_range=version["package_version_require_dbt_version"],
            package_id=version["package_id_from_path"],
        )
        versions.append(version)
        dbt_version_defined = package_version.is_require_dbt_version_defined()
        fusion_compatible_version: bool = (
            dbt_version_defined and package_version.is_require_dbt_version_fusion_compatible()
        )
        if package_version.is_version_explicitly_disallowed_on_fusion():
            fusion_compatible_version = False
        if fusion_compatible_version:
            fusion_compatible_versions.append(package_version.version)
            if not latest_fusion_version or package_version.version > latest_fusion_version:
                latest_fusion_version = package_version.version
            if not oldest_fusion_compatible_version or package_version.version < oldest_fusion_compatible_version:
                oldest_fusion_compatible_version = package_version.version
        elif not dbt_version_defined:
            unknown_compatibility_versions.append(package_version.version)
        elif dbt_version_defined and not fusion_compatible_version:
            fusion_incompatible_versions.append(package_version.version)
        if package_version.is_prerelease_version():
            if not latest_version_incl_prerelease or package_version.version > latest_version_incl_prerelease:
                latest_version_incl_prerelease = package_version.version
        elif not latest_version or package_version.version > latest_version:
            latest_version = package_version.version
    if latest_version_incl_prerelease is None:
        latest_version_incl_prerelease = latest_version
    if latest_version is None and latest_version_incl_prerelease:
        latest_version = latest_version_incl_prerelease
    assert latest_version is not None
    assert latest_version_incl_prerelease is not None
    latest_version_incl_prerelease = max(latest_version_incl_prerelease, latest_version)
    if not package_redirect_name and not package_redirect_namespace:
        assert package_latest_version_index_json == latest_version
        assert len(fusion_compatible_versions) + len(fusion_incompatible_versions) + len(
            unknown_compatibility_versions
        ) == len(versions)
        if latest_fusion_version:
            assert oldest_fusion_compatible_version is not None
            assert len(fusion_compatible_versions) > 0
        if not latest_fusion_version:
            assert oldest_fusion_compatible_version == None
            assert len(fusion_compatible_versions) == 0
    return {
        # "versions": versions,
        "latest_version": convert_version_spec_to_string(latest_version),
        "oldest_fusion_compatible_version": convert_version_spec_to_string(oldest_fusion_compatible_version),
        "latest_fusion_compatible_version": convert_version_spec_to_string(latest_fusion_version),
        "fusion_compatible_versions": [convert_version_spec_to_string(x) for x in fusion_compatible_versions],
        "fusion_incompatible_versions": [convert_version_spec_to_string(x) for x in fusion_incompatible_versions],
        "unknown_compatibility_versions": [convert_version_spec_to_string(x) for x in unknown_compatibility_versions],
        "package_latest_version_index_json": convert_version_spec_to_string(package_latest_version_index_json),
        "package_redirect_name": package_redirect_name,
        "package_redirect_namespace": package_redirect_namespace,
        "latest_version_incl_prerelease": convert_version_spec_to_string(latest_version_incl_prerelease),
    }


def get_versions(packages):
    # skip over renamed packages and come back once the new package is done
    renamed_packages: list[tuple[str, str]] = []
    packages_with_versions: dict[str, dict[str, Any]] = {}
    for package in packages:
        versions = get_versions_for_package(packages[package])
        if versions["package_redirect_name"] is not None or versions["package_redirect_namespace"] is not None:
            old_package_namespace = package.split("/")[0]
            old_package_name = package.split("/")[1]
            new_name = new_name_from_redirect(
                versions["package_redirect_name"],
                versions["package_redirect_namespace"],
                old_package_name,
                old_package_namespace,
            )
            renamed_packages.append((package, new_name))
            print(f"renamed package: {package}, {new_name}")
            continue
        else:
            packages_with_versions[package] = versions
    for package in renamed_packages:
        old_package_name = package[0]
        new_package_name = package[1]
        packages_with_versions[old_package_name] = packages_with_versions[new_package_name]
        redirect_namespace, redirect_name = new_package_name.split("/")
        packages_with_versions[old_package_name]["package_redirect_name"] = redirect_name
        packages_with_versions[old_package_name]["package_redirect_namespace"] = redirect_namespace
        packages_with_versions[old_package_name]["package_redirect_id"] = new_package_name
    assert len(packages_with_versions) == len(packages)
    return packages_with_versions


def write_dict_to_json(data: dict[str, Any], dest_dir: Path, *, indent: int = 2, sort_keys: bool = True) -> None:
    out_file = dest_dir / "fusion_version_compatibility_output.json"
    with out_file.open("w", encoding="utf-8") as fh:
        json.dump(data, fh, indent=indent, sort_keys=sort_keys, ensure_ascii=False)


def main():
    input: Path = Path.cwd() / "src" / "dbt_fusion_package_tools" / "scripts" / "output"
    data = read_package_output_json(input / "package_output.json")
    # check_package_names(data)
    packages_with_versions: dict[str, dict[str, Any]] = get_versions(data)
    print(f"Read {len(packages_with_versions)} packages from file")
    write_dict_to_json(packages_with_versions, input)
    print("Output written to fusion_version_compatibility_output.json")
    with open(
        Path.cwd() / "src" / "dbt_fusion_package_tools" / "fusion_version_compatibility_output.py",
        "w",
    ) as output_py_file:
        output_py_file.write(
            f"from typing import Any\n\nFUSION_VERSION_COMPATIBILITY_OUTPUT: dict[str, dict[str, Any]] = {packages_with_versions}"
        )


if __name__ == "__main__":
    main()



EXPLICIT_DISALLOW_ALL_VERSIONS: set[str] = set(
    [
        "Snowflake-Labs/dbt_constraints",
        "get-select/dbt_snowflake_query_tags",
        "data-mie/dbt_profiler",
        "dbt-labs/logging",
        "get-select/dbt_snowflake_monitoring",
        "dbt-labs/dbt_external_tables",
        "tnightengale/dbt_meta_testing",
        "yu-iskw/dbt_airflow_macros",
        "everpeace/dbt_models_metadata",
    ]
)

# https://docs.getdbt.com/docs/fusion/supported-features#package-support
EXPLICIT_ALLOW_ALL_VERSIONS: set[str] = set(
    [
        "fivetran/fivetran_utils",
        "fivetran/hubspot",
        "fivetran/linkedin",
        "fivetran/microsoft_ads",
        "fivetran/salesforce_formula_utils",
        "AxelThevenot/dbt_assertions",
        "Datavault-UK/automate_dv",
        "entechlog/dbt_snow_mask",
        "fivetran/ad_reporting",
        "fivetran/facebook_ads",
        "fivetran/fivetran_log",
        "fivetran/google_ads",
        "fivetran/jira",
        "fivetran/pendo",
        "fivetran/qualtrics",
        "fivetran/salesforce",
        "fivetran/social_media_reporting",
        "fivetran/zendesk",
        "godatadriven/dbt_date",
        "kristeligt-dagblad/dbt_ml",
        "metaplane/dbt_expectations",
        "Montreal-Analytics/snowflake_utils",
        "Snowflake-Labs/dbt_semantic_view",
        "dbt-labs/dbt_utils",
        "dbt-labs/audit_helper",
        "GJMcClintock/dbt_tld",
        "dbt-labs/codegen",
        "calogica/dbt_date",
    ]
)

# TODO: Currently this is used in scripts/get_fusion_compatible_versions
# to set compatibility when parsing the raw package files and also in
# DbtPackageVersion.is_version_explicitly_disallowed_on_fusion,
# but need to refine logic
EXPLICIT_DISALLOW_VERSIONS: dict[str, set[str]] = {
    # dbt_project_evaluator version 1.1.0 has compatible
    # require dbt version but actually has bug that makes
    # package incompatible until fixed in 1.1.2
    "dbt-labs/dbt_project_evaluator": set(["1.1.0", "1.1.1"]),
    "brooklyn-data/dbt_artifacts": set(
        [
            "0.1.0",
            "0.2.0",
            "0.2.1",
            "0.3.0",
            "0.4.0",
            "0.4.1",
            "0.4.2",
            "0.4.3",
            "0.4.4",
            "0.5.0",
            "0.6.0",
            "0.7.0",
            "0.8.0",
            "1.0.0",
            "1.1.0",
            "1.1.1",
            "1.1.2",
            "1.2.0",
            "2.0.0",
            "2.1.0",
            "2.1.1",
            "2.2.0",
            "2.2.1",
            "2.2.2",
            "2.2.3",
            "2.3.0",
            "2.4.0",
            "2.4.1",
            "2.4.2",
            "2.4.3",
            "2.5.0",
            "2.6.0",
            "2.6.1",
            "2.6.2",
            "2.6.3",
            "2.6.4",
            "2.7.0",
            "2.8.0",
            "2.9.0",
            "2.9.1",
            "2.9.2",
            "2.9.3",
        ]
    ),
}



from dataclasses import dataclass, field
from typing import Optional

from mashumaro.mixins.json import DataClassJSONMixin


@dataclass
class FusionCompatibility(DataClassJSONMixin):
    require_dbt_version_defined: Optional[bool] = None
    require_dbt_version_compatible: Optional[bool] = None
    fusion_parse: Optional[bool] = None
    dbt_verified: Optional[bool] = None


@dataclass
class FusionLogMessage(DataClassJSONMixin):
    body: str
    error_code: Optional[int] = None
    severity_text: Optional[str] = None
    """Original severity before user up/down-grade configuration applied."""
    original_severity_text: Optional[str] = None


@dataclass
class ParseConformanceLogOutput(DataClassJSONMixin):
    parse_exit_code: int = 0
    total_errors: int = 0
    total_warnings: int = 0
    errors: list[FusionLogMessage] = field(default_factory=list)
    warnings: list[FusionLogMessage] = field(default_factory=list)
    fusion_version: str = "unknown"


@dataclass
class FusionConformanceResult(DataClassJSONMixin):
    version: Optional[str] = None
    require_dbt_version_defined: Optional[bool] = None
    require_dbt_version_compatible: Optional[bool] = None
    parse_compatible: Optional[bool] = None
    parse_compatibility_result: Optional[ParseConformanceLogOutput] = None
    manually_verified_compatible: Optional[bool] = None
    manually_verified_incompatible: Optional[bool] = None
    download_failed: bool = False



"""Script to read conformance output and write to CSV.

Takes the conformance output that is generated by package_hub_fusion_compatibility
and converts it into CSV output for analysis.
Update OUTPUT_PATH, CONFORMANCE_OUTPUT_FILE_NAME, and FUSION_VERSION
Two CSVs are written to OUTPUT_PATH:
    "package_conformance_summary_v{fusion_version}_{run_date}.csv"
    "package_conformance_all_errors_v{fusion_version}_{run_date}.csv"
package_conformance_summary: one line per package version
package_conformance_all_errors: one line per error per package version
"""

import csv
import json
from collections import Counter, defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any

OUTPUT_PATH: Path = Path.cwd() / "src" / "dbt_fusion_package_tools" / "scripts" / "output"
CONFORMANCE_OUTPUT_FILE_NAME: str = "v101_conformance_output_20260128.json"
FUSION_VERSION = "101"


def reload_output_from_file(
    file_path: Path,
) -> defaultdict[str, list[dict[str, Any]]]:
    with file_path.open("r", encoding="utf-8") as fh:
        return json.load(fh)


def group_errors(error_code: str, error_message: str) -> str:
    if error_code == "102":
        return "dbt102: Deprecated test argument/config at top level"
    elif error_code == "1005" and error_message in (
        "'data-paths' cannot be specified in dbt_project.yml",
        "'source-paths' cannot be specified in dbt_project.yml",
    ):
        return "dbt1005: data-paths or source-paths in dbt_project.yml"
    elif error_code == "1005" and len(error_message) > 21 and error_message[0:21] == "Found duplicate model":
        return "dbt1005: Found duplicate model"
    else:
        remove_file_name = error_message.split("-->")
        split_on_backtick = (remove_file_name[0]).split("`")
        split_on_period = (split_on_backtick[0]).split(".")
        return f"dbt{error_code}: {split_on_period[0]}"


def main():
    output_path: Path = OUTPUT_PATH
    conformance_output_file_name: str = CONFORMANCE_OUTPUT_FILE_NAME
    fusion_version = FUSION_VERSION
    run_date: str = datetime.today().strftime("%Y%m%d")
    package_summary_file_name: str = f"package_conformance_summary_v{fusion_version}_{run_date}.csv"
    package_errors_file_name: str = f"package_conformance_all_errors_v{fusion_version}_{run_date}.csv"
    conformance_output = reload_output_from_file(output_path / conformance_output_file_name)
    package_summary = []
    unique_error_codes: Counter[str] = Counter()
    unique_warning_codes: Counter[str] = Counter()
    all_errors = []
    for package_name in conformance_output:
        package: dict[str, Any] = conformance_output[package_name]
        for version in package:
            package_version: dict[str, Any] = package[version]
            parse_errors: set[str] = set()
            parse_warnings: set[str] = set()
            parse_short_errors: set[str] = set()
            parse_compatibility_result = package_version.get("parse_compatibility_result") or {}
            if parse_compatibility_result.get("errors"):
                for error in parse_compatibility_result["errors"]:
                    parse_errors.add(str(error["error_code"]))
                    unique_error_codes[str(error["error_code"])] += 1
                    short_error = group_errors(str(error["error_code"]), error["body"])
                    parse_short_errors.add(short_error)
                    all_errors.append(
                        {
                            "package_name": package_name,
                            "package_version": version,
                            "manually_verified_compatible": package[version]["manually_verified_compatible"],
                            "manually_verified_incompatible": package[version]["manually_verified_incompatible"],
                            "parse_compatible": package[version]["parse_compatible"],
                            "parse_error_code": str(error["error_code"]),
                            "parse_error": error["body"] if len(error["body"]) < 200 else error["body"][:200],
                            "parse_exit_code": parse_compatibility_result.get("parse_exit_code"),
                            "parse_total_errors": parse_compatibility_result.get("total_errors"),
                            "parse_total_warnings": parse_compatibility_result.get("total_warnings"),
                            "require_dbt_version_compatible": package[version]["require_dbt_version_compatible"],
                            "require_dbt_version_defined": package[version]["require_dbt_version_defined"],
                            "fusion_version": f"v{fusion_version}",
                            "parse_error_grouped": short_error,
                        }
                    )
            if parse_compatibility_result.get("warnings"):
                for warning in parse_compatibility_result["warnings"]:
                    parse_warnings.add(str(warning["error_code"]))
                    unique_warning_codes[str(error["error_code"])] += 1
            sorted_parse_short_errors: list[str] = sorted([x for x in parse_short_errors])
            package_summary.append(
                {
                    "package_name": package_name,
                    "package_version": version,
                    "manually_verified_compatible": package[version]["manually_verified_compatible"],
                    "manually_verified_incompatible": package[version]["manually_verified_incompatible"],
                    "parse_errors": ",".join([f"dbt{x}" for x in sorted(parse_errors)])
                    if len(parse_errors) > 0
                    else "",
                    "parse_warnings": ",".join(sorted(parse_warnings)) if len(parse_warnings) > 0 else "",
                    "parse_exit_code": parse_compatibility_result.get("parse_exit_code"),
                    "parse_total_errors": parse_compatibility_result.get("total_errors"),
                    "parse_total_warnings": parse_compatibility_result.get("total_warnings"),
                    "parse_compatible": package[version]["parse_compatible"],
                    "require_dbt_version_compatible": package[version]["require_dbt_version_compatible"],
                    "require_dbt_version_defined": package[version]["require_dbt_version_defined"],
                    "fusion_version": f"v{fusion_version}",
                    "short_error_total": len(parse_short_errors),
                    "short_error_1": sorted_parse_short_errors[0] if len(sorted_parse_short_errors) > 0 else "",
                    "short_error_2": sorted_parse_short_errors[1] if len(sorted_parse_short_errors) > 1 else "",
                    "download_failed": package[version]["download_failed"],
                }
            )
    print(f"unique error codes: {unique_error_codes}")
    print(f"unique warning codes: {unique_warning_codes}")

    with open(output_path / package_summary_file_name, mode="w") as file:
        writer = csv.DictWriter(file, fieldnames=[field for field in package_summary[0]])
        writer.writeheader()
        writer.writerows(package_summary)
    print(f"Wrote {len(package_summary)} rows to {output_path / package_summary_file_name}")

    with open(output_path / package_errors_file_name, mode="w") as file:
        writer = csv.DictWriter(file, fieldnames=[field for field in all_errors[0]])
        writer.writeheader()
        writer.writerows(all_errors)
    print(f"Wrote {len(all_errors)} rows to {output_path / package_errors_file_name}")


if __name__ == "__main__":
    main()



from yaml.composer import Composer
from yaml.constructor import SafeConstructor
from yaml.parser import Parser
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner


class SafeConstructorWithOutput(SafeConstructor):
    def get_single_data(self):
        # Ensure that the stream contains a single document and construct it.
        node = self.get_single_node()
        if node is not None:
            document = self.construct_document(node)
            return document
        return None

    def get_single_data_and_document(self):
        node = self.get_single_node()
        if node is not None:
            return (node, self.construct_document(node))
        return None


class SafeLoaderWithOutput(Reader, Scanner, Parser, Composer, SafeConstructorWithOutput, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)


def load(stream, Loader):
    """Parse the first YAML document in a stream
    and produce the corresponding Python object.
    """
    loader = Loader(stream)

    try:
        return loader.get_single_data_and_document()
    finally:
        loader.dispose()


# Returns a tuple where the first entry is the raw mapping node and the second is the document
def safe_load(stream):
    return load(stream, SafeLoaderWithOutput)



from dataclasses import dataclass, field
from typing import Any, Optional

from rich.console import Console

from dbt_fusion_package_tools.dbt_package_version import (
    DbtPackageVersion,
)
from dbt_fusion_package_tools.fusion_version_compatibility_output import FUSION_VERSION_COMPATIBILITY_OUTPUT
from dbt_fusion_package_tools.manual_overrides import EXPLICIT_ALLOW_ALL_VERSIONS, EXPLICIT_DISALLOW_ALL_VERSIONS
from dbt_fusion_package_tools.upgrade_status import (
    PackageFusionCompatibilityState,
    PackageVersionFusionCompatibilityState,
)
from dbt_fusion_package_tools.version_utils import (
    VersionRange,
    VersionSpecifier,
    construct_version_list_from_raw,
    convert_optional_version_string_to_spec,
    convert_version_specifiers_to_range,
    convert_version_string_list_to_spec,
    get_version_specifiers,
    versions_compatible,
)

console = Console()
error_console = Console(stderr=True)


@dataclass
class DbtPackage:
    # name in project yml
    package_name: str
    # org/package_name used in deps and package hub
    package_id: str
    # version range specified in deps config (packages.yml)
    project_config_raw_version_specifier: Any
    project_config_version_range_list: list[str] = field(default_factory=list)
    project_config_version_range: Optional[VersionRange] = field(init=False)
    # package versions indexed by version string
    package_versions: dict[str, DbtPackageVersion] = field(default_factory=dict)
    installed_package_version: Optional[VersionSpecifier] = None
    latest_package_version: Optional[VersionSpecifier] = None
    latest_package_version_incl_prerelease: Optional[VersionSpecifier] = None
    # misc parameters from deps config
    git_url: Optional[str] = None
    opt_in_prerelease: bool = False
    local: bool = False
    tarball: bool = False
    git: bool = False
    private: bool = False
    # github repo from package hub
    github_repo_url: Optional[str] = None

    # fields from FUSION_VERSION_COMPATIBILITY_OUTPUT
    package_redirect_id: Optional[str] = None
    lowest_fusion_compatible_version: Optional[VersionSpecifier] = None
    fusion_compatible_versions: list[VersionSpecifier] = field(default_factory=list)
    fusion_incompatible_versions: list[VersionSpecifier] = field(default_factory=list)
    unknown_compatibility_versions: list[VersionSpecifier] = field(default_factory=list)

    # check compatibility of latest and installed versions when loading
    latest_version_fusion_compatibility: PackageVersionFusionCompatibilityState = (
        PackageVersionFusionCompatibilityState.UNKNOWN
    )
    installed_version_fusion_compatibility: PackageVersionFusionCompatibilityState = (
        PackageVersionFusionCompatibilityState.UNKNOWN
    )

    def merge_fusion_compatibility_output(self) -> bool:
        output = FUSION_VERSION_COMPATIBILITY_OUTPUT.get(self.package_id)
        if output is None:
            return False
        self.package_redirect_id = output.get("package_redirect_id")
        oldest_fusion_compatible_version = convert_optional_version_string_to_spec(
            output["oldest_fusion_compatible_version"]
        )
        fusion_compatible_versions = convert_version_string_list_to_spec(output["fusion_compatible_versions"])
        fusion_incompatible_versions = convert_version_string_list_to_spec(output["fusion_incompatible_versions"])
        unknown_compatibility_versions = convert_version_string_list_to_spec(output["unknown_compatibility_versions"])
        self.lowest_fusion_compatible_version = oldest_fusion_compatible_version
        self.fusion_compatible_versions = fusion_compatible_versions
        self.fusion_incompatible_versions = fusion_incompatible_versions
        self.unknown_compatibility_versions = unknown_compatibility_versions
        return True

    def __post_init__(self):
        try:
            if self.project_config_raw_version_specifier is not None:
                self.project_config_version_range_list = construct_version_list_from_raw(
                    self.project_config_raw_version_specifier
                )
            if self.project_config_version_range_list and len(self.project_config_version_range_list) > 0:
                version_specs: list[VersionSpecifier] = get_version_specifiers(self.project_config_version_range_list)
                self.project_config_version_range = convert_version_specifiers_to_range(version_specs)
            else:
                self.project_config_version_range = None
        except:
            self.project_config_version_range = None
            error_console.print("exception calculating config version range ")
        self.merge_fusion_compatibility_output()

    def add_package_version(self, new_package_version: DbtPackageVersion, installed=False, latest=False) -> bool:
        new_package_version.package_id = self.package_id
        if latest:
            self.latest_package_version = new_package_version.version
            self.latest_version_fusion_compatibility = new_package_version.get_fusion_compatibility_state()
        if new_package_version.package_version_str in self.package_versions:
            console.log(f"Package version {new_package_version.package_version_str} already exists in package versions")
            return False
        else:
            self.package_versions[new_package_version.package_version_str] = new_package_version
        if installed:
            self.installed_package_version = new_package_version.version
            self.installed_version_fusion_compatibility = new_package_version.get_fusion_compatibility_state()
        return True

    def set_latest_package_version(self, version_str: str, raw_require_dbt_version: Any = None):
        try:
            return self.add_package_version(
                DbtPackageVersion(
                    package_name=self.package_name,
                    package_version_str=version_str,
                    raw_require_dbt_version_range=raw_require_dbt_version,
                ),
                latest=True,
            )
        except:
            return False

    def is_public_package(self) -> bool:
        return not (self.git or self.tarball or self.local or self.private)

    def is_installed_version_fusion_compatible(self) -> PackageVersionFusionCompatibilityState:
        if self.installed_package_version is None:
            return PackageVersionFusionCompatibilityState.UNKNOWN
        else:
            installed_version_string = self.installed_package_version.to_version_string(skip_matcher=True)
            if installed_version_string not in self.package_versions:
                return PackageVersionFusionCompatibilityState.UNKNOWN
            else:
                return self.package_versions[installed_version_string].get_fusion_compatibility_state()

    def find_fusion_compatible_versions_in_requested_range(self) -> list[VersionSpecifier]:
        """Find package versions that are compatible with Fusion AND the version range specified in the project config.

        A project can upgrade to one of these version without updating their project's packages.yml.

        Returns:
            list[VersionSpecifier]: Fusion-compatible versions
        """
        if self.project_config_version_range is None:
            return []
        compatible_versions = []
        if self.fusion_compatible_versions is None or len(self.fusion_compatible_versions) == 0:
            return compatible_versions
        for version in self.fusion_compatible_versions:
            if versions_compatible(
                version, self.project_config_version_range.start, self.project_config_version_range.end
            ):
                compatible_versions.append(version)
        sorted_versions = sorted(compatible_versions)
        return sorted_versions

    def find_fusion_compatible_versions_above_requested_range(self) -> list[VersionSpecifier]:
        """Find package versions that are compatible with Fusion but NOT the version range specified in the project config.

        The project's packages.yml/dependencies.yml MUST be updated in order to upgrade to one of these version.

        Returns:
            list[VersionSpecifier]: Fusion-compatible versions
        """
        if self.project_config_version_range is None:
            return []
        compatible_versions = []
        if self.fusion_compatible_versions is None or len(self.fusion_compatible_versions) == 0:
            return compatible_versions
        for version in self.fusion_compatible_versions:
            if (
                not versions_compatible(
                    version,
                    self.project_config_version_range.start,
                    self.project_config_version_range.end,
                    # make sure we only count versions newer than the current version
                    # so we don't recommend downgrades
                )
                and version > self.project_config_version_range.start
            ):
                compatible_versions.append(version)
        sorted_versions = sorted(compatible_versions)
        return sorted_versions

    def find_fusion_incompatible_versions_in_requested_range(self) -> list[VersionSpecifier]:
        incompatible_versions = []
        if self.project_config_version_range is None:
            return []
        if self.fusion_incompatible_versions is None or len(self.fusion_incompatible_versions) == 0:
            return incompatible_versions
        for version in self.fusion_incompatible_versions:
            if versions_compatible(
                version, self.project_config_version_range.start, self.project_config_version_range.end
            ):
                incompatible_versions.append(version)
        sorted_versions = sorted(incompatible_versions)
        return sorted_versions

    def find_fusion_unknown_versions_in_requested_range(self) -> list[VersionSpecifier]:
        unknown_compatibility_versions = []
        if self.project_config_version_range is None:
            return []
        if self.unknown_compatibility_versions is None or len(self.unknown_compatibility_versions) == 0:
            return unknown_compatibility_versions
        for version in self.unknown_compatibility_versions:
            if versions_compatible(
                version, self.project_config_version_range.start, self.project_config_version_range.end
            ):
                unknown_compatibility_versions.append(version)
        sorted_versions = sorted(unknown_compatibility_versions)
        return sorted_versions

    def get_installed_package_version(self) -> str:
        if self.installed_package_version:
            return self.installed_package_version.to_version_string(skip_matcher=True)
        elif self.installed_package_version is None and self.project_config_version_range is not None:
            return self.project_config_version_range.start.to_version_string(skip_matcher=True)
        else:
            return "unknown"

    def get_package_fusion_compatibility_state(self) -> PackageFusionCompatibilityState:
        if not self.is_public_package():
            return PackageFusionCompatibilityState.UNKNOWN
        if self.package_id in EXPLICIT_DISALLOW_ALL_VERSIONS:
            return PackageFusionCompatibilityState.NO_VERSIONS_COMPATIBLE
        if self.package_id in EXPLICIT_ALLOW_ALL_VERSIONS:
            return PackageFusionCompatibilityState.ALL_VERSIONS_COMPATIBLE

        fusion_compatible_version_count = (
            0 if self.fusion_compatible_versions is None else len(self.fusion_compatible_versions)
        )
        fusion_incompatible_version_count = (
            0 if self.fusion_incompatible_versions is None else len(self.fusion_incompatible_versions)
        )
        unknown_compatibility_version_count = (
            0 if self.unknown_compatibility_versions is None else len(self.unknown_compatibility_versions)
        )
        total_version_count = (
            fusion_compatible_version_count + fusion_incompatible_version_count + unknown_compatibility_version_count
        )
        # cases where we can determine compatibility across all versions
        if total_version_count == 0:
            return PackageFusionCompatibilityState.UNKNOWN
        elif fusion_compatible_version_count == total_version_count:
            return PackageFusionCompatibilityState.ALL_VERSIONS_COMPATIBLE
        elif unknown_compatibility_version_count == total_version_count:
            return PackageFusionCompatibilityState.MISSING_COMPATIBILITY
        elif fusion_incompatible_version_count == total_version_count:
            return PackageFusionCompatibilityState.NO_VERSIONS_COMPATIBLE
        # case where we have to look at individual versions to determine compatibility
        elif total_version_count > 0 and fusion_compatible_version_count > 0:
            return PackageFusionCompatibilityState.SOME_VERSIONS_COMPATIBLE
        # fallback case where some versions have incompatible version and some have no version defined
        elif total_version_count > 0 and fusion_compatible_version_count == 0:
            return PackageFusionCompatibilityState.NO_VERSIONS_COMPATIBLE
        # hopefully nothing is left but if so
        else:
            return PackageFusionCompatibilityState.UNKNOWN



from typing import Optional


class SemverError(Exception):
    def __init__(self, msg: Optional[str] = None) -> None:
        self.msg = msg
        if msg is not None:
            super().__init__(msg)
        else:
            super().__init__()


class VersionsNotCompatibleError(SemverError):
    pass


class FusionBinaryNotAvailable(Exception):
    def __init__(self, message="Fusion binary not found on system, please install Fusion first"):
        self.message = message
        super().__init__(self.message)


class GitOperationError(Exception):
    """Custom exception for git operation failures"""

    pass



import json
import time
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
from requests import HTTPError


def _http_get_json(url: str, headers: Optional[Dict[str, str]] = None, timeout: int = 30) -> Any:
    try:
        resp = requests.get(url, headers=headers or {}, timeout=timeout)
        resp.raise_for_status()
        # requests already decodes JSON when using .json(), but in case
        # the content is not JSON, fall back to decoding manually.
        try:
            return resp.json()
        except ValueError:
            return json.loads(resp.text)
    except HTTPError:
        # re-raise HTTP errors to be handled by callers
        raise
    except requests.RequestException as exc:
        # Convert other request exceptions to a RuntimeError for clarity
        raise RuntimeError(f"Network error when fetching {url}: {exc}")


# Example package index path:
# data/packages/Aaron-Zhou/synapse_statistic/index.json
def is_package_index_file(file_path: str) -> bool:
    file_path_split = file_path.split("/")
    if len(file_path_split) != 5:
        return False
    return file_path_split[-1] == "index.json"


# Example package version path:
# data/packages/Aaron-Zhou/synapse_statistic/versions/v0.1.0.json
def is_package_version_file(file_path: str) -> bool:
    file_path_split = file_path.split("/")
    if len(file_path_split) != 6:
        return False
    return file_path_split[-2] == "versions"


# Example paths that resolve to Aaron-Zhou/synapse_statistic
# data/packages/Aaron-Zhou/synapse_statistic/index.json
# data/packages/Aaron-Zhou/synapse_statistic/versions/v0.1.0.json
def extract_package_id_from_path(file_path: str) -> str:
    file_path_split = file_path.split("/")
    if file_path_split[0] != "data" or file_path_split[1] != "packages" or len(file_path_split) < 4:
        return ""
    return f"{file_path_split[2]}/{file_path_split[3]}"


def clean_version(version_str: Optional[str]) -> str:
    """Remove leading 'v' or 'V' from version strings, if present."""
    if version_str is None:
        return ""
    elif version_str.startswith(("v", "V")):
        return version_str[1:]
    return version_str


def process_json(file_path: str, parsed_json: Any) -> dict[str, Any]:
    package_id = extract_package_id_from_path(file_path)
    if package_id == "":
        return {}
    if is_package_index_file(file_path):
        return {
            "package_id_from_path": package_id,
            "package_latest_version_index_json": clean_version(parsed_json.get("latest")),
            "package_name_index_json": parsed_json.get("name"),
            "package_namespace_index_json": parsed_json.get("namespace"),
            "package_redirect_name": parsed_json.get("redirectname"),
            "package_redirect_namespace": parsed_json.get("redirectnamespace"),
        }
    elif is_package_version_file(file_path):
        if "_source" in parsed_json:
            github_url = parsed_json["_source"].get("url", "")
        else:
            github_url = None
        if "downloads" in parsed_json:
            tarball_url = parsed_json["downloads"].get("tarball")
        else:
            tarball_url = None
        return {
            "package_id_from_path": package_id,
            "package_id_with_version": parsed_json.get("id"),
            "package_name_version_json": parsed_json.get("name"),
            "package_version_string": clean_version(parsed_json.get("version")),
            "package_version_require_dbt_version": parsed_json.get("require_dbt_version"),
            "package_version_github_url": github_url,
            "package_version_download_url": tarball_url,
        }
    else:
        return {}


def write_dict_to_json(data: Dict[str, Any], dest_dir: Path, *, indent: int = 2, sort_keys: bool = True) -> None:
    out_file = dest_dir / "package_output.json"
    with out_file.open("w", encoding="utf-8") as fh:
        json.dump(data, fh, indent=indent, sort_keys=sort_keys, ensure_ascii=False)


def download_package_jsons_from_hub_repo(
    owner: str = "dbt-labs",
    repo: str = "hub.getdbt.com",
    path: str = "data/packages",
    branch: Optional[str] = "master",
    github_token: Optional[str] = None,
    file_count_limit: int = 0,
) -> defaultdict[str, list[dict[str, Any]]]:
    """Download and parse all JSON files under `data/packages` in a GitHub repo.

    This function performs the following steps:
    - Discover the repository's default branch (if `branch` is not provided).
    - Fetch the git tree for the branch recursively and find all files under
      ``{path}`` that end with ``.json``.
    - Download each JSON file via the raw.githubusercontent.com URL and parse
      it into Python objects.

    Returns:
        A list of parsed JSON objects typed as ``PackageJSON``.

    Args:
        owner: GitHub repo owner (default: "dbt-labs").
        repo: GitHub repository name (default: "hub.getdbt.com").
        path: Path within the repo to search (default: "data/packages").
        branch: Branch name to use; if omitted the repository default branch is
            discovered via the GitHub API.
        github_token: Optional GitHub token to increase rate limits.
    """
    base_api = "https://api.github.com"
    headers: Dict[str, str] = {"User-Agent": "dbt-autofix-agent"}
    if github_token:
        headers["Authorization"] = f"token {github_token}"

    # 1) Find default branch if not provided
    if not branch:
        repo_url = f"{base_api}/repos/{owner}/{repo}"
        try:
            repo_info = _http_get_json(repo_url, headers=headers)
            branch = repo_info.get("default_branch")
        except Exception as exc:  # pragma: no cover - network error handling
            raise RuntimeError(f"Failed to get repo info for {owner}/{repo}: {exc}")
        if not branch:
            raise RuntimeError("Could not determine repository default branch")

    # 2) Get the git tree recursively
    tree_url = f"{base_api}/repos/{owner}/{repo}/git/trees/{branch}?recursive=1"
    try:
        tree_json = _http_get_json(tree_url, headers=headers)
    except Exception as exc:  # pragma: no cover - network error handling
        raise RuntimeError(f"Failed to fetch git tree for {owner}/{repo}@{branch}: {exc}")

    if "tree" not in tree_json:
        raise RuntimeError("Unexpected response from GitHub API when fetching tree")

    files: List[Dict[str, Any]] = []
    prefix = path.rstrip("/") + "/"
    for entry in tree_json["tree"]:
        # entry has keys: path, mode, type (blob/tree), sha, url
        if entry.get("type") != "blob":
            continue
        p = entry.get("path", "")
        if p.startswith(prefix) and p.endswith(".json"):
            files.append(entry)
        if file_count_limit > 0 and len(files) >= file_count_limit:
            break

    packages: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)
    if not files:
        # No files found; return empty list rather than error.
        return packages

    # 3) Download each JSON using raw.githubusercontent.com
    for entry in files:
        file_path = entry["path"]
        raw_url = f"https://raw.githubusercontent.com/{owner}/{repo}/{branch}/{file_path}"
        # Use simple GET; raw.githubusercontent does not require auth for public repos.
        try:
            parsed = _http_get_json(raw_url, headers={"User-Agent": headers["User-Agent"]})
            output = process_json(file_path, parsed)
            if output != {}:
                packages[output["package_id_from_path"]].append(output)
            time.sleep(1)
        except Exception as exc:  # network/file parsing issues
            warnings.warn(f"Failed to fetch/parse {file_path}: {exc}")
            time.sleep(5)

    return packages


def read_json_from_local_hub_repo(path: str, file_count_limit: int = 0):
    """Read JSON files from a local copy of the hub repo and return a
    defaultdict mapping package_id -> list[parsed outputs].

    The `path` argument may be either:
      - the repository root (so files are found under `data/packages/...`),
      - or the `data/packages` directory itself, or
      - a single JSON file path.

    Behavior mirrors `download_package_jsons_from_hub_repo` where possible:
      - JSON files are found recursively
      - each file is parsed and passed to `process_json(file_path, parsed_json)`
      - parsing/IO errors are warned and skipped
    """
    base = Path(path)
    packages: defaultdict[str, list[dict[str, Any]]] = defaultdict(list)

    if not base.exists():
        warnings.warn(f"Path does not exist: {path}")
        return packages

    # Collect JSON files
    json_files: List[Path]
    if base.is_file():
        if base.suffix.lower() == ".json":
            json_files = [base]
        else:
            return packages
    else:
        json_files = sorted(base.rglob("*.json"), key=lambda p: str(p))

    if file_count_limit > 0:
        json_files = json_files[:file_count_limit]

    if not json_files:
        return packages

    for file in json_files:
        try:
            with file.open("r", encoding="utf-8") as fh:
                parsed = json.load(fh)

            # Try to produce a repo-style path like 'data/packages/...'
            file_path: str
            parts = list(file.parts)
            if "data" in parts:
                idx = parts.index("data")
                file_path = Path(*parts[idx:]).as_posix()
            else:
                try:
                    # prefer path relative to provided base
                    rel = file.relative_to(base)
                    file_path = rel.as_posix()
                except Exception:
                    file_path = file.as_posix()

            # If the user passed the `data/packages` directory itself,
            # ensure returned path still starts with 'data/packages'
            if not file_path.startswith("data/packages") and base.name == "packages" and base.parent.name == "data":
                rel = file.relative_to(base)
                file_path = Path("data") / "packages" / rel
                file_path = file_path.as_posix()

            output = process_json(file_path, parsed)
            if output != {}:
                packages[output["package_id_from_path"]].append(output)
        except Exception as exc:
            warnings.warn(f"Failed to read/parse {file}: {exc}")

    return packages


def reload_packages_from_file(
    file_path: Path,
) -> defaultdict[str, list[dict[str, Any]]]:
    with file_path.open("r", encoding="utf-8") as fh:
        return json.load(fh)


def main():
    file_count_limit = 0
    results = read_json_from_local_hub_repo(path="~/workplace/hub.getdbt.com", file_count_limit=file_count_limit)
    print(f"Downloaded {len(results)} packages from hub.getdbt.com")
    output_path: Path = Path.cwd() / "src" / "dbt_fusion_package_tools" / "scripts" / "output"
    write_dict_to_json(results, output_path)
    print(f"Output written to {output_path / 'package_output.json'}")
    reload_packages = reload_packages_from_file(output_path / "package_output.json")
    print(f"Reloaded {len(reload_packages)} packages from file")


if __name__ == "__main__":
    main()



from pdm.backend.hooks.version import SCMVersion


def format_version(version: SCMVersion) -> str:
    if version.distance is None:
        return str(version.version)
    else:
        return f"{version.version}.post{version.distance}"



"""Runs dbt parse in Fusion to identify parse erors in packages."""

import json
import os
import re
import subprocess
import tarfile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import requests
from dbtlabs.proto.public.v1.events.fusion.invocation.invocation_pb2 import Invocation
from dbtlabs.proto.public.v1.events.fusion.log.log_pb2 import LogMessage
from google.protobuf import json_format
from rich.console import Console

from dbt_fusion_package_tools.compatibility import (
    FusionConformanceResult,
    FusionLogMessage,
    ParseConformanceLogOutput,
)
from dbt_fusion_package_tools.dbt_package_version import DbtPackageVersion
from dbt_fusion_package_tools.exceptions import FusionBinaryNotAvailable
from dbt_fusion_package_tools.yaml.loader import safe_load

console = Console()
error_console = Console(stderr=True)

"""Cleans model paths displayed in Fusion error messages.

Example:
    Original: ../../../../../../private/var/folders/xf/r3bz6sxx0m57976q8nb2d1fh0000gp/T/tmp0bnpllg5/models/examples/basic_example/basic_example_d_site.yml:10:9
    Cleaned: models/examples/basic_example/basic_example_d_site.yml:10:9
"""
_ERROR_PATH_REGEX = re.compile(r"(?:\.\./)*")


def construct_download_url_from_latest(
    latest_package_version_download_url: str, target_version_download_url: str
) -> str:
    version_download_tag: str = target_version_download_url.split("/")[-1]
    latest_version_url: list[str] = latest_package_version_download_url.split("/")[:-1]
    latest_version_url.append(version_download_tag)
    return "/".join(latest_version_url)


def download_tarball_and_run_conformance(
    package_name: str,
    package_id: str,
    package_version_str: str,
    package_version_download_url: str,
    latest_package_version_download_url: Optional[str],
    fusion_binary: Optional[str] = None,
) -> Optional[FusionConformanceResult]:
    with TemporaryDirectory() as tmpdir:
        # download tarball from version json
        tar_path: Optional[Path] = None
        # track exceptions across both checks
        exceptions: list[str] = []
        # use explicitly provided version first
        try:
            # Download the tarball
            response = requests.get(package_version_download_url, stream=True)
            response.raise_for_status()

            # Save to a temporary file
            tar_path = Path(tmpdir) / "archive.tar.gz"
            with open(tar_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        except requests.exceptions.HTTPError as http_error:
            exceptions.append(
                f"{http_error.request.url}: {http_error.response.status_code}, {http_error.response.reason}"
            )
        except Exception as other_error:
            exceptions.append(f"Error when downloading tarball: {other_error}")
        # if that errors or doesn't exist, construct from the latest version
        if not tar_path and latest_package_version_download_url:
            constructed_url: str = construct_download_url_from_latest(
                latest_package_version_download_url, package_version_download_url
            )
            try:
                # Download the tarball
                response = requests.get(constructed_url, stream=True)
                response.raise_for_status()

                # Save to a temporary file
                tar_path = Path(tmpdir) / "archive.tar.gz"
                with open(tar_path, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
            except requests.exceptions.HTTPError as http_error:
                exceptions.append(
                    f"{http_error.request.url}: {http_error.response.status_code}, {http_error.response.reason}"
                )
            except Exception as other_error:
                exceptions.append(f"Error when downloading tarball: {other_error}")
        # if still no download, error
        if not tar_path:
            console.log(f"Could not download {package_name} {package_version_str}")
            for exception in exceptions:
                console.log(exception)
            return FusionConformanceResult(version=package_version_str, download_failed=True)

        # if we do have a file, extract the archive
        try:
            extract_dir = Path(tmpdir) / "extracted"
            extract_dir.mkdir(parents=True, exist_ok=True)

            with tarfile.open(tar_path, "r:gz") as tar:
                for entry in tar:
                    if os.path.isabs(entry.name) or ".." in entry.name:
                        raise ValueError("Illegal tar archive entry")
                    tar.extract(entry, extract_dir)

            # Clean up the tar file
            tar_path.unlink()

            # Check that only 1 directory is inside
            tar_contents = os.listdir(extract_dir)
            if len(tar_contents) != 1:
                console.log("Error downloading tar")
            extracted_package = extract_dir / tar_contents[0]
        except Exception as e:
            console.log(f"Error when extracting tarball: {e}")
            return

        # run conformance if possible
        try:
            console.log(f"Running parse conformance for {package_id} version {package_version_str}")
            return run_conformance_for_version(
                extracted_package, package_name, package_version_str, package_id, fusion_binary=fusion_binary
            )
        except Exception as e:
            console.log(f"Error when running conformance: {e}")
            return


def run_conformance_for_version(
    path, package_name, tag_version, package_id, fusion_binary=None
) -> Optional[FusionConformanceResult]:
    result = FusionConformanceResult(version=tag_version, download_failed=False)
    # check require dbt version
    try:
        dbt_project_yml = safe_load((Path(f"{path}/dbt_project.yml")).read_text()) or (
            {},
            {},
        )
        require_dbt_version_string = dbt_project_yml[1].get("require-dbt-version")
    except Exception as e:
        error_console.log(f"dbt_project.yml load failed for {package_id} {tag_version}: {e}")
        return
    # try to add profile to suppress warning about profiles
    if "profile" not in dbt_project_yml[1]:
        console.log("Adding profile to dbt project")
        try:
            with open(Path(f"{path}/dbt_project.yml"), "a") as f:
                f.write("\nprofile: test_schema_compat\n")
        except Exception as e:
            error_console.log(f"failed when adding profile to dbt_project.yml for {package_id} {tag_version}: {e}")
    new_version: DbtPackageVersion = DbtPackageVersion(
        package_name,
        tag_version,
        package_id=package_id,
        raw_require_dbt_version_range=require_dbt_version_string,
    )
    parse_conformance = check_fusion_schema_compatibility(
        Path(path), fusion_binary=fusion_binary, show_fusion_output=True
    )
    result.require_dbt_version_defined = new_version.is_require_dbt_version_defined()
    if result.require_dbt_version_defined:
        result.require_dbt_version_compatible = new_version.is_require_dbt_version_fusion_compatible()
    result.manually_verified_compatible = new_version.is_explicitly_allowed_on_fusion()
    result.manually_verified_incompatible = new_version.is_explicitly_disallowed_on_fusion()
    if parse_conformance:
        result.parse_compatible = parse_conformance.parse_exit_code == 0
        result.parse_compatibility_result = parse_conformance
    return result


def check_binary_name(binary_name: str) -> bool:
    try:
        subprocess.run(
            [
                binary_name,
                "--version",
            ],
            capture_output=True,
            timeout=60,
            check=True,
        )
        return True
    # indicates that the binary name is not found
    except FileNotFoundError:
        return False
    # indicates that an error occurred when running command
    except subprocess.CalledProcessError as process_error:
        error_console.log(
            f"CalledProcessError: {binary_name} --version exited with return code {process_error.returncode}"
        )
        error_console.log(process_error.stderr)
        return False
    except Exception as other_error:
        error_console.log(f"{other_error}: An unknown exception occured when running {binary_name} --version")
        return False


def check_fusion_version(binary_name: str) -> Optional[str]:
    try:
        version_result = subprocess.run(
            [
                binary_name,
                "--version",
            ],
            check=False,
            capture_output=True,
            timeout=60,
            text=True,
        )
        if "dbt-fusion" in version_result.stdout:
            return version_result.stdout.split()[-1]
    except Exception as e:
        error_console.log(f"{e}: An unknown error occurred when checking dbt version")
    return


def find_fusion_binary(custom_name: Optional[str] = None) -> Optional[str]:
    possible_binary_names: list[str] = ["dbtf", "dbt"] if custom_name is None else [custom_name]

    binary_names_found: set[str] = set()
    # test each name
    for binary_name in possible_binary_names:
        # first check if name exists at all
        binary_exists = check_binary_name(binary_name)
        if binary_exists:
            binary_names_found.add(binary_name)

    if len(binary_names_found) == 0:
        error_console.log("No fusion binaries found on system path, please install first")
        return None

    # now check version returned by each and use first one
    # don't need exception handling here because previous step already did it
    for valid_binary_name in binary_names_found:
        version_result = check_fusion_version(valid_binary_name)
        if version_result:
            return valid_binary_name

    # if we got to the end, then no fusion version has been found
    error_console.log(f"Could not find Fusion binary, latest version output is {version_result}")
    return None


def parse_log_output(
    output: str, exit_code: int, repo_path: Path = Path.cwd(), fusion_version: str = "unknown"
) -> ParseConformanceLogOutput:
    log_output = [json.loads(x) for x in output.splitlines()]
    result = ParseConformanceLogOutput(parse_exit_code=exit_code, fusion_version=fusion_version)
    temp_path = str(repo_path)
    for line in log_output:
        if line.get("event_type") == "v1.public.events.fusion.log.LogMessage":
            severity_text = line.get("severity_text")
            # remove the tempdir path from the log message
            body_no_path = (str(line.get("body"))).replace(temp_path, "")
            body_no_path = body_no_path.replace("private/", "")
            body = _ERROR_PATH_REGEX.sub("", body_no_path)
            log_message = LogMessage()
            json_format.ParseDict(line.get("attributes"), log_message, ignore_unknown_fields=True)
            fusion_log_message = FusionLogMessage(
                body=body,
                severity_text=str(severity_text),
                error_code=log_message.code,
                original_severity_text=log_message.original_severity_text,
            )
            if severity_text == "ERROR":
                result.errors.append(fusion_log_message)
            elif severity_text == "WARNING":
                result.warnings.append(fusion_log_message)
        elif (
            line.get("record_type") == "SpanEnd"
            and line.get("event_type") == "v1.public.events.fusion.invocation.Invocation"
        ):
            invocation = Invocation()
            json_format.ParseDict(line.get("attributes"), invocation, ignore_unknown_fields=True)
            result.total_errors += invocation.metrics.total_errors
            result.total_warnings += invocation.metrics.total_warnings

    return result


def check_fusion_schema_compatibility(
    repo_path: Path = Path.cwd(),
    fusion_binary: Optional[str] = None,
    show_fusion_output=True,
) -> Optional[ParseConformanceLogOutput]:
    """Check if a dbt package is fusion schema compatible by running 'dbtf parse'.

    Args:
        fusion_binary_name: name of a valid Fusion binary
        repo_path: Path to the dbt package repository

    Returns:
        True if fusion compatible (dbtf parse exits with code 0), False otherwise
    """
    # Add a test profiles.yml to the current directory
    profiles_path = repo_path / Path("profiles.yml")
    try:
        with open(profiles_path, "a") as f:
            f.write(
                "\n"
                "test_schema_compat:\n"
                "  target: dev\n"
                "  outputs:\n"
                "    dev:\n"
                "      type: postgres\n"
                "      host: localhost\n"
                "      port: 5432\n"
                "      user: postgres\n"
                "      password: postgres\n"
                "      dbname: postgres\n"
                "      schema: public\n"
            )

        # Ensure the `_DBT_FUSION_STRICT_MODE` is set (this will ensure fusion errors on schema violations)
        os.environ["_DBT_FUSION_STRICT_MODE"] = "1"

        # Find correct name for Fusion binary if none provided
        fusion_binary_name: Optional[str] = fusion_binary
        if fusion_binary_name is None:
            fusion_binary_name = find_fusion_binary()
        # If still no valid name, return
        if fusion_binary_name is None:
            raise FusionBinaryNotAvailable()

        # Get the Fusion version
        fusion_version: Optional[str] = check_fusion_version(fusion_binary_name)
        if fusion_version is None:
            raise FusionBinaryNotAvailable()

        try:
            # Run dbt deps to install package dependencies
            if show_fusion_output:
                console.log("\n\nRunning dbt deps", style="green")
            deps_result = subprocess.run(
                [
                    fusion_binary_name,
                    "deps",
                    "--profile",
                    "test_schema_compat",
                    "--project-dir",
                    str(repo_path),
                    "--log-format",
                    "otel",
                ],
                check=False,
                text=True,
                capture_output=True,
                timeout=60,
            )
            deps_output = parse_log_output(
                deps_result.stdout, deps_result.returncode, repo_path, fusion_version=fusion_version
            )
            if deps_result.returncode != 0:
                error_console.log("dbt deps returned errors")
                error_console.log(deps_output)

            # Now try parse
            if show_fusion_output:
                console.log("\n\nRunning dbt parse", style="green")
            parse_result = subprocess.run(
                [
                    fusion_binary_name,
                    "parse",
                    "--profile",
                    "test_schema_compat",
                    "--project-dir",
                    str(repo_path),
                    "--log-format",
                    "otel",
                ],
                check=False,
                text=True,
                capture_output=True,
                timeout=60,
            )
            parse_output = parse_log_output(
                parse_result.stdout, parse_result.returncode, repo_path, fusion_version=fusion_version
            )
            if parse_result.returncode != 0:
                error_console.log("dbt parse returned errors")
                error_console.log(parse_output)
                if len(deps_output.errors) > 0:
                    parse_output.errors.extend(deps_output.errors)
        except Exception as e:
            error_console.log(f"{e}: An unknown error occurred when running dbt parse")
            return

        # Return True if exit code is 0 (success)
        is_compatible = parse_result.returncode == 0

        if show_fusion_output:
            if is_compatible:
                console.log(f"Package at {repo_path} is fusion schema compatible")
            else:
                console.log(f"Package at {repo_path} is not fusion schema compatible")

        # Clean up deps
        if show_fusion_output:
            console.log("\n\nRunning dbt clean", style="green")
        subprocess.run(
            [
                fusion_binary_name,
                "clean",
                "--profile",
                "test_schema_compat",
                "--project-dir",
                str(repo_path),
                "--log-format",
                "otel",
            ],
            check=False,
            timeout=60,
            text=True,
            capture_output=True,
        )
        # Remove the test profile
        os.remove(profiles_path)

        return parse_output

    except Exception as e:
        error_console.log(f"Error checking fusion compatibility for {repo_path}: {e!s}")
        try:
            os.remove(profiles_path)
        except Exception:
            pass
        return


def main():
    check_fusion_schema_compatibility(repo_path=Path.cwd())


if __name__ == "__main__":
    main()



import csv
from pathlib import Path
from typing import Any

from dbt_fusion_package_tools.fusion_version_compatibility_output import FUSION_VERSION_COMPATIBILITY_OUTPUT


def main():
    package_summary = []
    for package_name in FUSION_VERSION_COMPATIBILITY_OUTPUT:
        package: dict[str, Any] = FUSION_VERSION_COMPATIBILITY_OUTPUT[package_name]
        package_summary.append(
            {
                "package_name": package_name,
                "fusion_compatible_version": len(package["fusion_compatible_versions"]),
                "fusion_incompatible_versions": len(package["fusion_incompatible_versions"]),
                "unknown_versions": len(package["unknown_compatibility_versions"]),
                "redirect": package["package_redirect_name"] is not None
                or package["package_redirect_namespace"] is not None,
                "latest_is_prerelease": package["latest_version"] != package["latest_version_incl_prerelease"],
                "latest_version": package["latest_version"],
                "oldest_fusion_compatible_version": package["oldest_fusion_compatible_version"],
                "latest_fusion_compatible_version": package["latest_fusion_compatible_version"],
            }
        )
    field_names = [field for field in package_summary[0]]
    output_file = Path.cwd() / "src" / "dbt_fusion_package_tools" / "scripts" / "output" / "packages.csv"
    with open(output_file, mode="w") as file:
        writer = csv.DictWriter(file, fieldnames=field_names)
        writer.writeheader()
        writer.writerows(package_summary)
    print(f"Wrote {len(package_summary)} rows to packages.csv")


if __name__ == "__main__":
    main()



from enum import Enum


class PackageVersionFusionCompatibilityState(str, Enum):
    """String enum for Fusion compatibility of a specific version of a package."""

    NO_DBT_VERSION_RANGE = "require-dbt-version is not defined"
    DBT_VERSION_RANGE_EXCLUDES_2_0 = "require-dbt-version excludes version 2.0"
    DBT_VERSION_RANGE_INCLUDES_2_0 = "require-dbt-version includes version 2.0"
    EXPLICIT_ALLOW = "Version has been verified by dbt as Fusion-compatible even though its declared require-dbt-version may not include 2.0"
    EXPLICIT_DISALLOW = "Version has been verified by dbt as incompatible with Fusion"
    UNKNOWN = "Version state unknown"


class PackageFusionCompatibilityState(str, Enum):
    """String enum for Fusion compatibility at the package level."""

    ALL_VERSIONS_COMPATIBLE = "All package versions are Fusion compatible"
    SOME_VERSIONS_COMPATIBLE = "A subset of package versions are Fusion compatible"
    NO_VERSIONS_COMPATIBLE = "No versions are Fusion compatible"
    MISSING_COMPATIBILITY = "All package versions are missing require dbt version"
    UNKNOWN = "Package version state unknown"


class PackageVersionUpgradeType(str, Enum):
    """String enum for package upgrade types"""

    NO_UPGRADE_REQUIRED = "Package is already compatible with Fusion"
    UPGRADE_AVAILABLE = "Package has Fusion-compatible version available"
    PUBLIC_PACKAGE_MISSING_FUSION_ELIGIBILITY = (
        "Public package has not defined require-dbt-version so Fusion eligibility cannot be determined"
    )
    PUBLIC_PACKAGE_NOT_COMPATIBLE_WITH_FUSION = "Public package is not compatible with Fusion"
    PUBLIC_PACKAGE_FUSION_COMPATIBLE_VERSION_EXCEEDS_PROJECT_CONFIG = (
        "Public package has Fusion-compatible version that is outside the project's requested version range"
    )
    PRIVATE_PACKAGE_MISSING_REQUIRE_DBT_VERSION = "Private package requires a compatible require-dbt-version (>=2.0.0, <3.0.0) to be available on Fusion. https://docs.getdbt.com/reference/project-configs/require-dbt-version"
    UNKNOWN = "Package's Fusion eligibility unknown"
